using System;
using System.Runtime.InteropServices;
using System.Security;
using GameKnight.Core;

using Collections = System.Collections.Generic;
using Drawing = System.Drawing;
using Forms = System.Windows.Forms;
using IO = System.IO;
using Reflection = System.Reflection;
using Resources = System.Resources;
using Threading = System.Threading;

using Direct3D = SlimDX.Direct3D;
using Direct3D9 = SlimDX.Direct3D9;

namespace GameKnight.Win32
{

	public sealed class Application : IInterface
	{
		
		[StructLayout(LayoutKind.Sequential)]
		private struct Message
		{

			public IntPtr Handle;

			public IntPtr WindowMessage;

			public IntPtr wParam;

			public IntPtr lParam;

			public uint Time;

			public Drawing.Point Point;

		}
		
		[SuppressUnmanagedCodeSecurity]
		[DllImport( "User32.dll", CharSet=CharSet.Auto )]
		private static extern bool PeekMessage( out Message Message, IntPtr Handle, uint MessageFilterMinimum, uint MessageFilterMaximum, uint Flags );

		private bool Active = true;

		private IConsole Console
		{
			get
			{
				return ConsoleProperty;
			}
			set
			{
				ConsoleProperty = value;
			}
		}

		private IConsole ConsoleProperty = null;
		
		private readonly ConfigFile ConfigFile = null;

		private static readonly string Directory = ( IO.Directory.GetCurrentDirectory() + "\\" );
		
		private readonly Direct3D9.Device DrawDevice = null;
		
		private readonly Forms.Panel DrawPanel = new Forms.Panel();
			
		private readonly Forms.Form Form = new Forms.Form();
 
		private int FPS = 0;
		
		private readonly Forms.StatusBarPanel FPS_Panel = new Forms.StatusBarPanel();
		
		private readonly Forms.Timer FrameUpdateTimer = new Forms.Timer();
		
        private static bool Idle
		{
			get
			{
				Message Message;
				return ( PeekMessage( out Message, IntPtr.Zero, 0, 0, 0 ) == false );
			}
		}

        private InputDevice InputDevice = null;
		 
        private readonly Forms.MenuStrip MenuStrip = new Forms.MenuStrip();

        private readonly Forms.Timer MessageTimer = new Forms.Timer();
        
      	public IOptions Options
		{
			get
			{
				return ConfigFile;
			}
		}
      			
      	private readonly Direct3D9.PresentParameters PresentParameters = new Direct3D9.PresentParameters();
		
		private int StateIndex = 1;

        private readonly Forms.StatusBarPanel StatePanel = new Forms.StatusBarPanel();
        
        private readonly Forms.StatusBar StatusBar = new Forms.StatusBar();
           
        private readonly Direct3D9.Texture Texture = null;
        
        public ITimer Timer
		{
			get
			{
				return TimerProperty;
			}
		}
        
        private readonly Timer TimerProperty = new Timer();
		
        private static readonly CustomVertex[] VertexArray =
		{
			new CustomVertex( 0.0f, 0.0f, 1.0f, 0.0f ),
			new CustomVertex( 256.0f, 0.0f, 0.0f, 0.0f ),
			new CustomVertex( 0.0f, 256.0f, 1.0f, 1.0f ),
			new CustomVertex( 256.0f, 256.0f, 0.0f, 1.0f ),
		};
		
        public Application()
		{
        	ConfigFile = new ConfigFile( Directory + "GameBoy.ini", true );	
        	Resources.ResourceManager ResourceManager = new Resources.ResourceManager( "GameKnight.Win32.Win32.Resources", Reflection.Assembly.GetExecutingAssembly() );
			Form.Text = "GameKnight";
			Form.Icon = (Drawing.Icon)ResourceManager.GetObject( "Icon" );
			Form.Activated += FormActivatedEvent;
			Form.LocationChanged += FormActivatedEvent;
			Form.SizeChanged += FormActivatedEvent;
			Form.Deactivate += FormDeactivateEvent;
			Form.Move += FormDeactivateEvent;
			Form.Resize += FormDeactivateEvent;
			Form.Resize += FormResizeEvent;
			Form.KeyDown += FormKeyDownEvent;
			Form.KeyUp += FormKeyUpEvent;
			Form.Controls.Add( MenuStrip );
			Form.Controls.Add( DrawPanel );
			Form.Controls.Add( StatusBar );
			Form.ClientSize = new Drawing.Size( 320, 288 + MenuStrip.Height + StatusBar.Height );
			MenuStrip.Items.Add( new FileMenu( LoadMenuItemEvent, ResetMenuItemEvent, ExitMenuItemEvent ) );
			MenuStrip.Items.Add( new OptionsMenu( InputMenuItemEvent, FrameSkipMenuItemEvent ) );
			StatusBar.ShowPanels = true;
			StatusBar.Panels.Add( StatePanel );
			StatusBar.Panels.Add( FPS_Panel );		
			//DrawDevice = new DrawDevice( DisplayPanel );
			//InputDevice = new InputDevice( Form );
			/*DrawDevice = new DirectDraw.Device( DirectDraw.CreateFlags.Default );
			DrawDevice.SetCooperativeLevel( DisplayPanel, DirectDraw.CooperativeLevelFlags.Normal );
			DirectDraw.SurfaceCaps SurfaceCaps = new DirectDraw.SurfaceCaps();
			SurfaceCaps.PrimarySurface = true;
			PrimarySurface = new DirectDraw.Surface( new DirectDraw.SurfaceDescription( SurfaceCaps ), DrawDevice );
			PrimarySurface.Clipper = new DirectDraw.Clipper( DrawDevice );
			PrimarySurface.Clipper.Window = DisplayPanel;*/	
			Direct3D9.Direct3D.Initialize();
			PresentParameters.BackBufferCount = 1;
			PresentParameters.EnableAutoDepthStencil = false;
			PresentParameters.PresentationInterval = Direct3D9.PresentInterval.Immediate;
			PresentParameters.Windowed = true;
			PresentParameters.DeviceWindowHandle = DrawPanel.Handle;
			PresentParameters.SwapEffect = Direct3D9.SwapEffect.Discard;
			DrawDevice = new Direct3D9.Device( 0, Direct3D9.DeviceType.Hardware, PresentParameters.DeviceWindowHandle, ( Direct3D9.CreateFlags.MixedVertexProcessing | Direct3D9.CreateFlags.FpuPreserve ), PresentParameters );
			DrawDevice.SetTransform( Direct3D9.TransformState.World, SlimDX.Matrix.Identity );
			DrawDevice.SetTransform( Direct3D9.TransformState.View, SlimDX.Matrix.Identity );
			DrawDevice.SetSamplerState( 0, Direct3D9.SamplerState.MipFilter, Direct3D9.TextureFilter.None );
			DrawDevice.SetSamplerState( 0, Direct3D9.SamplerState.MinFilter, Direct3D9.TextureFilter.None );
			DrawDevice.SetSamplerState( 0, Direct3D9.SamplerState.MagFilter, Direct3D9.TextureFilter.None );
			DrawDevice.SetRenderState( Direct3D9.RenderState.AlphaTestEnable, false );
			DrawDevice.SetRenderState( Direct3D9.RenderState.Lighting, false );
			DrawDevice.SetRenderState( Direct3D9.RenderState.ZEnable, false );
			DrawDevice.SetRenderState( Direct3D9.RenderState.CullMode, Direct3D9.Cull.Clockwise );
			Texture = new Direct3D9.Texture( DrawDevice, 256, 256, 1, Direct3D9.Usage.None, Direct3D9.Format.A8R8G8B8, Direct3D9.Pool.Managed );
			MessageTimer.Interval = 4000;
			MessageTimer.Tick += new EventHandler( MessageTimerTickEvent );
			FrameUpdateTimer.Interval = 1000;
			FrameUpdateTimer.Tick += new EventHandler( FrameUpdateTimerTickEvent );
			FrameUpdateTimer.Start();
			Form.Show();
			return;
		}
        
        public void Error( object Sender, string Message )
        {
        	return;
        }

		private void ExitMenuItemEvent( object Sender, EventArgs Arguments )
		{
			Form.Close();
			return;	
		}

		private void FormActivatedEvent( object Sender, EventArgs Arguments )
		{
			Active = true;
			TimerProperty.Resume();
			return;
		}

		private void FormDeactivateEvent( object Sender, EventArgs Arguments )
		{
			Active = false;
			TimerProperty.Pause();
			return;
		}
		
		private void FormKeyDownEvent( object Sender, Forms.KeyEventArgs Arguments )
		{
			switch( Arguments.KeyCode )
			{	
				case Forms.Keys.D0:
				case Forms.Keys.D1:
				case Forms.Keys.D2:
				case Forms.Keys.D3:
				case Forms.Keys.D4:
				case Forms.Keys.D5:
				case Forms.Keys.D6:
				case Forms.Keys.D7:
				case Forms.Keys.D8:
				case Forms.Keys.D9:
					if( Console != null )
					{
						StateIndex = ((int)Arguments.KeyCode) - ((int)Forms.Keys.D0);
						StatePanel.Text = String.Format( "Save state slot: {0}", StateIndex );
						MessageTimer.Stop();
						MessageTimer.Start();
					}
					break;
				/*case Forms.Keys.F5:
					if( Console != null )
					{
						if( SaveState( String.Format( "{0}.sv{1}", Console.Cartridge.Path, StateIndex ) ) == false )
  						{
							StatePanel.Text = String.Format( "Error saving state: {0}", StateIndex );
                    	}
                    	else
                    	{
							StatePanel.Text = String.Format( "Saved state: {0}", StateIndex );
                    	}
                    	MessageTimer.Stop();
						MessageTimer.Start();
					}
					break;
				case Forms.Keys.F8:
					if( Console != null )
					{
                        if( LoadState( String.Format( "{0}.sv{1}", Console.Cartridge.Path, StateIndex ) ) == false )
  						{
							StatePanel.Text = String.Format( "Error loading state: {0}", StateIndex );
                    	}
                    	else
                    	{
							StatePanel.Text = String.Format( "Loaded state: {0}", StateIndex );
                    	}
                    	MessageTimer.Stop();
						MessageTimer.Start();
                    }
					break;*/
				default:
					if( Console != null )
					{
						InputDevice[Arguments.KeyCode] = true;
					}
					break;
			}
			return;
		}
		
		private void FormKeyUpEvent( object Sender, Forms.KeyEventArgs Arguments )
		{
			switch( Arguments.KeyCode )
			{	
				case Forms.Keys.D0:
				case Forms.Keys.D1:
				case Forms.Keys.D2:
				case Forms.Keys.D3:
				case Forms.Keys.D4:
				case Forms.Keys.D5:
				case Forms.Keys.D6:
				case Forms.Keys.D7:
				case Forms.Keys.D8:
				case Forms.Keys.D9:
					break;
				case Forms.Keys.F5:
					break;
				case Forms.Keys.F8:
					break;
				default:
					if( Console != null )
					{
						InputDevice[Arguments.KeyCode] = false;
					}
					break;
			}
			return;
		}

		private void FormResizeEvent( object Sender, EventArgs Arguments )
		{
			if( Form.ClientRectangle.Width > 0 )
			{
				StatePanel.Width = Form.ClientRectangle.Width - 100;
				FPS_Panel.Width = Form.ClientRectangle.Width - StatePanel.Width;
				DrawPanel.Bounds = Drawing.Rectangle.FromLTRB( 0, MenuStrip.Bottom, Form.ClientRectangle.Width, StatusBar.Top );
			}
			return;
		}

		private void FrameSkipMenuItemEvent( object Sender, EventArgs Arguments )
		{
			if( Console != null )
			{
				ConfigFile.Set( "VideoProcessor", "FrameSkip", Sender.ToString() );
			}
			return;	
		}
		
		private void FrameUpdateTimerTickEvent( object Sender, EventArgs Arguments )
		{
			FPS_Panel.Text = ( ( Console == null ) ? "" : ( Console.VideoProcessor.FPS.ToString() + " FPS" ) );
			FPS = 0;
			return;
		}
		
		private void IdleEvent( object Sender, EventArgs Arguments )
		{
			while( Idle )
			{
				if( ( Active ) && ( Console != null ) )
				{
					Console.Update();
				}
				else
				{
					Threading.Thread.Sleep( 50 );
				}
			}
			return;
		}
		
		private void InputMenuItemEvent( object Sender, EventArgs Arguments )
		{
			if( Console != null )
			{
				new ControllerDialog( Console, InputDevice ).ShowDialog( Form );
			}
			return;	
		}
		
		/*private bool LoadState( string Path )
		{
			IO.BinaryReader BinaryReader = null;
			try
			{
				BinaryReader = new IO.BinaryReader( IO.File.OpenRead( Path ) );
			}
			catch
			{
				return false;
			}
			if( Console.SerializationArray == null )
			{
				BinaryReader.Close();
				return false;	
			}
			if( BinaryReader.ReadString() != "GMKT" )
			{
				return false;
			}
			foreach( ISerializable CurrentObject in Console.SerializationArray )
			{
				if( CurrentObject.Load( BinaryReader ) == false )
				{
					BinaryReader.Close();
					return false;	
				}
			}
			BinaryReader.Close();
			return true;
		}*/	
			
		private void LoadMenuItemEvent( object Sender, EventArgs Arguments )
		{
			Forms.OpenFileDialog OpenFileDialog = new Forms.OpenFileDialog();
			OpenFileDialog.Filter = "GameBoy ROM(*.gb,*.sgb,*.cgb,*.gbc)|*.gb;*.sgb;*.cgb;*.gbc|All Files(*.*)|*.*;";
			OpenFileDialog.ReadOnlyChecked = true;
			OpenFileDialog.ShowReadOnly = true;
			OpenFileDialog.Title = "Load ROM";
			if( OpenFileDialog.ShowDialog() == Forms.DialogResult.OK )
			{
				IO.FileInfo FileInfo = new IO.FileInfo( OpenFileDialog.FileName );
				IO.BinaryReader BinaryReader = new IO.BinaryReader( IO.File.OpenRead( FileInfo.FullName ) );
				switch( FileInfo.Extension.ToLower() )
				{
					case ".gb":
					case ".sgb":
					case ".cgb":
					case ".gbc":
						if( ( Console is GameBoy.Console ) == false )
						{
							Console = new GameBoy.Console( this );
						}
						Console.Load( new InputStream( FileInfo.FullName ) );
						Form.ClientSize = new Drawing.Size( ( Console.VideoProcessor.ScreenSize.Width * 2 ), ( Console.VideoProcessor.ScreenSize.Height * 2 ) + MenuStrip.Height + StatusBar.Height );
						DrawDevice.SetTransform( Direct3D9.TransformState.Projection, SlimDX.Matrix.OrthoOffCenterLH( 256.0f, ( Console.VideoProcessor.ScreenSize.Width - 64 ), Console.VideoProcessor.ScreenSize.Height, 0.0f, 0.1f, 100.0f ) );
						FPS_Panel.Text = "";		
						InputDevice = new InputDevice( Console.ControllerArray[0], ConfigFile );
						Form.Text = "GameKnight - " + FileInfo.Name;
						Timer.Reset();
						break;
				}
				BinaryReader.Close();
			}
			return;	
		}
		
		private void MessageTimerTickEvent( object Sender, EventArgs Arguments )
		{
			StatePanel.Text = "";
			MessageTimer.Stop();
			return;
		}
		
		public void Present( uint[] ColorArray )
		{
			Direct3D.LockedRect LockedRect = Texture.LockRectangle( 0, new Drawing.Rectangle( 0, 0, 160, 144 ), Direct3D9.LockFlags.None );
			unsafe
			{
				fixed( void* Pointer = &ColorArray[0] )
				{
					LockedRect.Data.Write( new IntPtr( Pointer ), ( 256 * Console.VideoProcessor.ScreenSize.Width * 4 ) );
				}
			}
			LockedRect.Data.Close();
			Texture.UnlockRectangle( 0 );
			DrawDevice.Clear( Direct3D9.ClearFlags.Target, Drawing.Color.Black, 1.0f, 0 );
			DrawDevice.BeginScene();
			DrawDevice.VertexFormat = CustomVertex.Format;
			DrawDevice.SetTexture( 0, Texture );
			DrawDevice.DrawUserPrimitives( Direct3D9.PrimitiveType.TriangleStrip, 0, 2, VertexArray );
			DrawDevice.EndScene();
			DrawDevice.Present();
			FPS++;
			return;
		}
		
		private void ResetMenuItemEvent( object Sender, EventArgs Arguments )
		{
			if( Console != null )
			{
				Console.Reset();
				InputDevice = new InputDevice( Console.ControllerArray[0], ConfigFile );
			}
			return;	
		}
				
		/*private bool SaveState( string Path )
		{
			IO.BinaryWriter BinaryWriter = null;
			try
			{
				BinaryWriter = new IO.BinaryWriter( IO.File.OpenWrite( Path ) );
			}
			catch
			{
				return false;
			}
			if( Console.SerializationArray == null )
			{
				BinaryWriter.Close();
				return false;	
			}
			BinaryWriter.Write( "GMKT" );
			foreach( ISerializable CurrentObject in Console.SerializationArray )
			{
				if( CurrentObject.Save( BinaryWriter ) == false )
				{
					BinaryWriter.Close();
					return false;	
				}
			}
			BinaryWriter.Close();
			return true;
		}*/

		[STAThread]
		private static void Main( string[] ArgumentArray )
		{
			Forms.Application.EnableVisualStyles();
			Application Application = new Application();
			Forms.Application.Idle += new EventHandler( Application.IdleEvent );
			Forms.Application.Run( Application.Form );
			return;
		}

	}

}
